# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

__author__    = 'ihaveamac'
__copyright__ = 'Copyright (c) 2017-2019 Ian Burgwin'
__license__   = 'MIT'
__version__   = '1.5'
